# Test file with Unicode imports
# These should fail or be handled appropriately

import ö
from ä import something
import café
from naïve import something_else
import 日本語
from Москва import test
from москва2 import test2

